﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.DataPreprocessing {
  [Item("Data Grid", "Represents a data grid.")]
  [StorableClass]
  public class DataGridContent : PreprocessingContent, IStringConvertibleMatrix, IViewShortcut {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Table; }
    }

    public int Rows {
      get { return PreprocessingData.Rows; }
      set { }
    }

    public int Columns {
      get { return PreprocessingData.Columns; }
      set { }
    }

    public IEnumerable<string> ColumnNames {
      get { return PreprocessingData.VariableNames; }
      set { }
    }

    public IEnumerable<string> RowNames {
      get { return Enumerable.Range(1, Rows).Select(n => n.ToString()); }
      set { throw new NotSupportedException(); }
    }

    public bool SortableView {
      get { return true; }
      set { throw new NotSupportedException(); }
    }

    public bool ReadOnly {
      get { return false; }
    }

    public IDictionary<int, IList<int>> Selection {
      get { return PreprocessingData.Selection; }
      set { PreprocessingData.Selection = value; }
    }

    #region Constructor, Cloning & Persistence
    public DataGridContent(IFilteredPreprocessingData preprocessingData)
      : base(preprocessingData) {
    }

    public DataGridContent(DataGridContent original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DataGridContent(this, cloner);
    }

    [StorableConstructor]
    protected DataGridContent(bool deserializing)
      : base(deserializing) { }
    #endregion

    public void DeleteRows(IEnumerable<int> rows) {
      PreprocessingData.DeleteRows(rows);
    }

    public void DeleteColumn(int column) {
      PreprocessingData.DeleteColumn(column);
    }

    public bool Validate(string value, out string errorMessage, int columnIndex) {
      return PreprocessingData.Validate(value, out errorMessage, columnIndex);
    }

    public string GetValue(int rowIndex, int columnIndex) {
      return PreprocessingData.GetCellAsString(columnIndex, rowIndex);
    }

    public bool SetValue(string value, int rowIndex, int columnIndex) {
      return PreprocessingData.SetValue(value, columnIndex, rowIndex);
    }

    public event DataPreprocessingChangedEventHandler Changed {
      add { PreprocessingData.Changed += value; }
      remove { PreprocessingData.Changed -= value; }
    }

    #region unused stuff/not implemented but necessary due to IStringConvertibleMatrix
#pragma warning disable 0067
    // Is not used since DataGridContentView overrides dataGridView_CellValidating and uses 
    // DataGridLogic#Validate(string value, out string errorMessage, int columnIndex)
    public bool Validate(string value, out string errorMessage) {
      errorMessage = string.Empty;
      return true;
    }

    public event EventHandler ColumnsChanged;
    public event EventHandler RowsChanged;
    public event EventHandler ColumnNamesChanged;
    public event EventHandler RowNamesChanged;
    public event EventHandler SortableViewChanged;
    public event EventHandler<EventArgs<int, int>> ItemChanged;
    public event EventHandler Reset;
#pragma warning restore 0067
    #endregion

    #region Manipulations
    public void ReplaceIndicesByMean(IDictionary<int, IList<int>> cells, bool considerSelection = false) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          PreprocessingData.DataColumns[column.Key].TypeSwitch(
            c => {
              var mean = c.GetMean(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = mean;
            },
            dateTimeAction: c => {
              var mean = c.GetMean(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = mean;
            });
        }
      });
    }

    public void ReplaceIndicesByMedianValue(IDictionary<int, IList<int>> cells, bool considerSelection = false) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          PreprocessingData.DataColumns[column.Key].TypeSwitch(
            c => {
              var median = c.GetMedian(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = median;
            },
            c => {
              var median = c.GetMedian(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = median;
            },
            c => {
              var median = c.GetMedian(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = median;
            });
        }
      });
    }

    public void ReplaceIndicesByMode(IDictionary<int, IList<int>> cells, bool considerSelection = false) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          PreprocessingData.DataColumns[column.Key].TypeSwitch(
            c => {
              var mode = c.GetMode(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = mode;
            },
            c => {
              var mode = c.GetMode(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = mode;
            },
            c => {
              var mode = c.GetMode(considerSelection ? column.Value : null);
              foreach (var index in column.Value) c[index] = mode;
            });
        }
      });
    }

    public void ReplaceIndicesByRandomValue(IDictionary<int, IList<int>> cells, bool considerSelection = false) {
      var rand = new FastRandom();
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          PreprocessingData.DataColumns[column.Key].TypeSwitch(
            c => {
              double min = c.GetMin(considerSelection ? column.Value : null);
              double max = c.GetMax(considerSelection ? column.Value : null);
              double range = max - min;
              foreach (var index in column.Value) c[index] = min + rand.NextDouble() * range;
            },
            dateTimeAction: c => {
              var min = c.GetMin(considerSelection ? column.Value : null);
              var max = c.GetMax(considerSelection ? column.Value : null);
              double range = (max - min).TotalSeconds;
              foreach (var index in column.Value) c[index] = min + TimeSpan.FromSeconds(rand.NextDouble() * range);
            });
        }
      });
    }

    public void ReplaceIndicesByString(IDictionary<int, IList<int>> cells, string value) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          foreach (var rowIdx in column.Value) {
            PreprocessingData.DataColumns[column.Key].SetValue(value, rowIdx);
          }
        }
      });
    }


    public void ReplaceIndicesByLinearInterpolationOfNeighbours(IDictionary<int, IList<int>> cells) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          IList<Tuple<int, int>> startEndings = GetStartAndEndingsForInterpolation(column);
          foreach (var tuple in startEndings) {
            Interpolate(column, tuple.Item1, tuple.Item2);
          }
        }
      });
    }

    private List<Tuple<int, int>> GetStartAndEndingsForInterpolation(KeyValuePair<int, IList<int>> column) {
      var startEndings = new List<Tuple<int, int>>();
      var rowIndices = column.Value.OrderBy(x => x).ToList();
      var count = rowIndices.Count;
      int start = int.MinValue;
      for (int i = 0; i < count; ++i) {
        if (start == int.MinValue) {
          start = IndexOfPrevPresentValue(column.Key, rowIndices[i]);
        }
        if (i + 1 == count || (i + 1 < count && rowIndices[i + 1] - rowIndices[i] > 1)) {
          int next = IndexOfNextPresentValue(column.Key, rowIndices[i]);
          if (start > 0 && next < PreprocessingData.Rows) {
            startEndings.Add(new Tuple<int, int>(start, next));
          }
          start = int.MinValue;
        }
      }
      return startEndings;
    }

    private void Interpolate(KeyValuePair<int, IList<int>> column, int prevIndex, int nextIndex) {
      int valuesToInterpolate = nextIndex - prevIndex;

      PreprocessingData.DataColumns[column.Key].TypeSwitch(
        c => {
          double prev = c[prevIndex];
          double next = c[nextIndex];
          double interpolationStep = (next - prev) / valuesToInterpolate;
          for (int i = prevIndex; i < nextIndex; i++) c[i] = prev + (interpolationStep * (i - prevIndex));
        },
        dateTimeAction: c => {
          var prev = c[prevIndex];
          var next = c[nextIndex];
          double interpolationStep = (next - prev).TotalSeconds / valuesToInterpolate;
          for (int i = prevIndex; i < nextIndex; i++) c[i] = prev.AddSeconds(interpolationStep * (i - prevIndex));
        }
      );
    }

    private int IndexOfPrevPresentValue(int columnIndex, int start) {
      int index = start - 1;
      while (index >= 0 && PreprocessingData.IsCellEmpty(columnIndex, index))
        index--;
      return index;
    }

    private int IndexOfNextPresentValue(int columnIndex, int start) {
      int index = start + 1;
      while (index < PreprocessingData.Rows && PreprocessingData.IsCellEmpty(columnIndex, index))
        index++;
      return index;
    }

    public void Shuffle(bool shuffleRangesSeparately) {
      var random = new FastRandom();

      if (shuffleRangesSeparately) {
        var ranges = new[] { PreprocessingData.TestPartition, PreprocessingData.TrainingPartition };
        PreprocessingData.InTransaction(() => {
          // process all given ranges - e.g. TrainingPartition, TestPartition
          foreach (var range in ranges) {
            var indices = Enumerable.Range(0, PreprocessingData.Rows).ToArray();
            var shuffledIndices = Enumerable.Range(range.Start, range.Size).Shuffle(random).ToArray();
            for (int i = range.Start, j = 0; i < range.End; i++, j++)
              indices[i] = shuffledIndices[j];

            ReOrderToIndices(indices);
          }
        });

      } else {
        PreprocessingData.InTransaction(() => {
          var indices = Enumerable.Range(0, PreprocessingData.Rows).ToArray();
          indices.ShuffleInPlace(random);
          ReOrderToIndices(indices);
        });
      }
    }

    public void ReOrderToIndices(int[] indices) {
      PreprocessingData.InTransaction(() => {
        foreach (var column in PreprocessingData.DataColumns) {
          column.TypeSwitch(
            c => {
              if (indices.Length != c.Values.Count) throw new InvalidOperationException("The number of provided indices does not match the values.");
              var originalData = new List<double>(c.Values);
              for (int i = 0; i < indices.Length; i++) c[i] = originalData[indices[i]];
            },
            c => {
              if (indices.Length != c.Values.Count) throw new InvalidOperationException("The number of provided indices does not match the values.");
              var originalData = new List<string>(c.Values);
              for (int i = 0; i < indices.Length; i++) c[i] = originalData[indices[i]];
            },
            c => {
              if (indices.Length != c.Values.Count) throw new InvalidOperationException("The number of provided indices does not match the values.");
              var originalData = new List<DateTime>(c.Values);
              for (int i = 0; i < indices.Length; i++) c[i] = originalData[indices[i]];
            });
        }
      });
    }
    #endregion
  }
}
